/*
 *   Copyright Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

/**
 * @file   bitmap_free.c
 * @author Ilias Pliotas, Ioannis Koutras
 * @date   September 2012
 * @brief  free() implementation for bitmap-organized raw blocks
 */

#include "bitmap/bitmap.h"
#include "bitmap/bitmap_rb.h"
#include "bitmap/bitmap_other.h"

#include "trace.h"

#ifdef WITH_ALLOCATOR_STATS
#include "dmmlib/dmmlib.h"
#endif /* WITH_ALLOCATOR_STATS */

/** Frees the memory block inside of a specific bitmap-organized raw block.
 * @param raw_block The pointer of the bitmap raw block.
 * @param ptr       The pointer of the memory block to be freed.
 */
void bitmap_free(bitmap_rb_t *raw_block, void *ptr) {
    chunk_header_t *chunk_header;
    size_t cells_used, cell_no; 
    BMAP_EL_TYPE *bmap_p;

    chunk_header = (chunk_header_t *)((char *)ptr - CHUNK_HDR_SIZE);

    cells_used = chunk_header->num_of_cells;

#ifndef REQUEST_SIZE_INFO
    TRACE_1("dmmlib - free - free'ing %zu bytes from bitmap raw block %p\n",
            cells_used * raw_block->bytes_per_cell, (void *)raw_block);
#else /* REQUEST_SIZE_INFO */
    TRACE_1("dmmlib - free - free'ing %zu bytes from bitmap raw block %p, out of "
            "which %zu were used by the application\n",
            cells_used * raw_block->bytes_per_cell,
            (void *)raw_block,
            chunk_header->requested_size);
#endif /* REQUEST_SIZE_INFO */

#ifdef WITH_ALLOCATOR_STATS
    systemallocator.dmm_stats.total_mem_allocated -= cells_used *
        raw_block->bytes_per_cell;
    TRACE_1("dmmlib - global allocated memory: %zu bytes\n",
            systemallocator.dmm_stats.total_mem_allocated);
#ifdef REQUEST_SIZE_INFO
    systemallocator.dmm_stats.total_mem_requested -= chunk_header->requested_size;
    TRACE_1("dmmlib - global requested memory: %zu bytes\n",
            systemallocator.dmm_stats.total_mem_requested);
#endif /* REQUEST_SIZE_INFO */
    systemallocator.dmm_stats.live_objects--;
    systemallocator.dmm_stats.num_free++;
#endif /* WITH_ALLOCATOR_STATS */

    cell_no = (size_t) ((char *)chunk_header -
            ((char *)raw_block + sizeof(bitmap_rb_t) + 
             raw_block->elements * BMAP_EL_SIZE))
        / raw_block->bytes_per_cell;

    bmap_p = (BMAP_EL_TYPE *)((char *)raw_block + sizeof(bitmap_rb_t));

    size_t mask_counter = cells_used;
    size_t mask_start = cell_no % BMAP_EL_SIZE_BITS + 1;
    size_t vector_index = cell_no / BMAP_EL_SIZE_BITS;

    while(mask_counter != 0) {
        if(mask_counter > BMAP_EL_SIZE_BITS) {
            bmap_p[vector_index] |= make_bit_mask(mask_start, 
                    BMAP_EL_SIZE_BITS - mask_start + 1);
            mask_counter -= BMAP_EL_SIZE_BITS - mask_start + 1;
            vector_index++;
            mask_start = 1;
        } else {
            bmap_p[vector_index] |= make_bit_mask(mask_start, 
                    mask_counter);
            mask_counter = 0;
        }
    }
}
