/*
 *   Copyright Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

/**
 * @file   free.c
 * @author Ioannis Koutras (joko@microlab.ntua.gr)
 * @date   September 2012
 *
 * @brief  Implementation free() call.
 */

#include "dmmlib/dmmlib.h"

#include <stdbool.h>

#ifdef BITMAP_RB_ONLY
#include "bitmap/bitmap.h"
#include "bitmap/bitmap_rb.h"
#endif /* BITMAP_RB_ONLY */

#ifdef FL_RB_ONLY
#include "freelist/freelist.h"
#include "freelist/freelist_rb.h"
#endif /* FL_RB_ONLY */

#include "release_memory.h"

#include "trace.h"

void free(void *ptr) {
    raw_block_header_t *current_raw_block;
    bool found;

    if(ptr == NULL) {
        return;
    }

    found = false;

    current_raw_block = systemallocator.raw_block_head;
    while(current_raw_block) {
        if(((char *)ptr > (char *)current_raw_block) &&
                ((char *)ptr < (char *)(current_raw_block) +
                 current_raw_block->size)) {
            found = true;
            break;
        }
        current_raw_block = current_raw_block->next_raw_block;
    }

    if(found == true) {

#ifdef BITMAP_RB_ONLY
        bitmap_rb_t
#endif /* BITMAP_RB_ONLY */
#ifdef FL_RB_ONLY
        freelist_rb_t
#endif /* FL_RB_ONLY */
            *encapsulated_rb = 
#ifdef BITMAP_RB_ONLY
                (bitmap_rb_t *)
#endif /* BITMAP_RB_ONLY */
#ifdef FL_RB_ONLY
                (freelist_rb_t *)
#endif /* FL_RB_ONLY */
                ((char *)current_raw_block + sizeof(raw_block_header_t));

#ifdef HAVE_LOCKS
        pthread_mutex_lock(&current_raw_block->mutex);
#endif /* HAVE_LOCKS */
        dmmlib_free(encapsulated_rb, ptr);
#ifdef HAVE_LOCKS
        pthread_mutex_unlock(&current_raw_block->mutex);
#endif /* HAVE_LOCKS */

    } else { // It has to be a BIGBLOCK, just munmap it

        current_raw_block = (raw_block_header_t *)((char *)ptr -
                sizeof(raw_block_header_t));
        TRACE_1("dmmlib - free - free'ing %zu bytes from raw block %p\n",
                current_raw_block->size, (void *)current_raw_block);

#ifdef WITH_ALLOCATOR_STATS
        systemallocator.dmm_stats.total_mem_allocated -=
            current_raw_block->size;
        TRACE_1("dmmlib - global allocated memory: %zu bytes\n",
                systemallocator.dmm_stats.total_mem_allocated);
#ifdef REQUEST_SIZE_INFO
        systemallocator.dmm_stats.total_mem_requested -=
            current_raw_block->size - sizeof(raw_block_header_t);
        TRACE_1("dmmlib - global requested memory: %zu bytes\n",
                systemallocator.dmm_stats.total_mem_requested);
#endif /* REQUEST_SIZE_INFO */
        systemallocator.dmm_stats.live_objects--;
        systemallocator.dmm_stats.num_free++;
#endif /* WITH_ALLOCATOR_STATS */

        release_memory(current_raw_block);
    }
}
