/*
 *   Copyright Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

/**
 * @file   freelist_free.c
 * @author Ioannis Koutras
 * @brief  free() implementation for freelist-organised raw blocks
 */

#include "freelist/freelist.h"

#include "freelist/block_header_funcs.h"
#include "freelist/linked_lists/add_block.h"
#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
#include "freelist/coalesce.h"
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */

#include "trace.h"

#ifdef WITH_ALLOCATOR_STATS
#include "dmmlib/dmmlib.h"
#endif /* WITH_ALLOCATOR_STATS */

/** Frees the memory block inside of a specific free-list organized raw block.
 * @param raw_block The pointer of the freelist raw block.
 * @param ptr       The pointer of the memory block to be freed.
 */
void freelist_free(freelist_rb_t *raw_block, void *ptr) {

    block_header_t *block = get_header(ptr);

#ifndef REQUEST_SIZE_INFO
    TRACE_1("dmmlib - free - free'ing %zu bytes from freelist raw block %p\n",
            get_size(block), (void *)raw_block);
#else /* REQUEST_SIZE_INFO */
    TRACE_1("dmmlib - free - free'ing %zu bytes from freelist raw block %p, out of "
            "which %zu were used by the application\n",
            get_size(block),
            (void *)raw_block,
            get_requested_size(block));
#endif /* REQUEST_SIZE_INFO */

    // Memory stats get updated here in case the space gets coalesced with its
    // free neighbors.
#ifdef WITH_ALLOCATOR_STATS
    systemallocator.dmm_stats.total_mem_allocated -=
        get_size(block) + HEADER_SIZE;
    TRACE_1("dmmlib - global allocated memory: %zu bytes\n",
            systemallocator.dmm_stats.total_mem_allocated);
#ifdef REQUEST_SIZE_INFO
    systemallocator.dmm_stats.total_mem_requested -=
        get_requested_size(block);
    TRACE_1("dmmlib - global requested memory: %zu bytes\n",
            systemallocator.dmm_stats.total_mem_requested);
#endif /* REQUEST_SIZE_INFO */
    systemallocator.dmm_stats.live_objects--;
    systemallocator.dmm_stats.num_free++;
#endif /* WITH_ALLOCATOR_STATS */

#if defined (COALESCING_FIXED) || defined (COALESCING_VARIABLE)
    coalesce(raw_block, block);
#else
    mark_free(raw_block, block);
    add_block(raw_block, block);
#endif /* COALESCING_FIXED || COALESCING_VARIABLE */

}
